"""
    authorization_django
    ~~~~~~~~~~~~~~~~~~~~

    Authorization middleware that uses JWTs for authentication.

    The following settings are used by the middleware, and can be configured in
    your ``settings.py`` in the ``DATAPUNT_AUTHZ`` dictionary.

    .. tabularcolumns:: |p{6.5cm}|p{8.5cm}|

    ================================= =========================================
    ``JWT_SECRET_KEY``                (Required) Your JWT signing key
    ``JWT_ALGORITHM``                 (Required) Algorithm to use for the JWT
                                      message authentication code (MAC)
    ``LOGGER_NAME``                   Name of the logger. (Default =
                                      ``authorization_django``)
    ``LOGGER_LEVEL``                  Log level. Will be overwritten if running
                                      debug mode. (Default = ``INFO``)
    ``LOGGER_FORMAT``                 Log format
    ``LOGGER_FORMAT_DEBUG``           Log format for messages in debug mode
"""
from .middleware import authorization_middleware

__all__ = (
    'authorization_middleware',
)



"""
    authorization_middleware.config
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
import functools
import logging
import types

from . import jwks
from django.conf import settings as django_settings

# A sentinel object for required settings
_required = object()

# The Django settings key
_settings_key = 'DATAPUNT_AUTHZ'

# A list of all available settings, with default values
_available_settings = {
    'JWKS': _required,
    'MIN_SCOPE': tuple(),
    'FORCED_ANONYMOUS_ROUTES': tuple(),
    'LOGGER_NAME': __name__,
    'LOGGER_LEVEL': logging.INFO,
    'LOGGER_FORMAT': '[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
    'LOGGER_FORMAT_DEBUG': (
        '-' * 80 + '\n' +
        '%(levelname)s in %(module)s [%(pathname)s:%(lineno)d]:\n' +
        '%(message)s\n' +
        '-' * 80
    )
}

# Validator functions and error messages
_settings_rectifiers = {
    'JWKS': {
        'func': lambda s: jwks.load(s),
        'errmsg': 'Must provide a valid JWKSet. See RFC 7517 and 7518 for details.'
    },
    'FORCED_ANONYMOUS_ROUTES': {
        'func': lambda s: type(s) in {list, tuple, set} and s,
        'errmsg': 'FORCED_ANONYMOUS_ROUTES must be a list, tuple or set'
    }
}

# Preprocessing: the set of all available configuration keys
_available_settings_keys = set(_available_settings.keys())

# Preprocessing: the set of all required configuration keys
_required_settings_keys = {
    key for key, setting in _available_settings.items() if setting is _required
}


class AuthzConfigurationError(Exception):
    """ Error for missing / invalid configuration"""


def _rectify(settings):
    """ Rectify (and validate) the given settings using the functions in
    :data:`_settings_rectifiers`.
    """
    for key, rectifier in _settings_rectifiers.items():
        try:
            new_value = rectifier['func'](settings[key])
            if new_value is False:
                raise AuthzConfigurationError(
                    'Error validating {}->{}: {}'.format(
                        _settings_key, key, rectifier['errmsg']))
            settings[key] = new_value
        except:
            raise AuthzConfigurationError(
                'Error validating {}->{}: {}'.format(
                    _settings_key, key, rectifier['errmsg']))


@functools.lru_cache(maxsize=1)
def settings():
    """ Fetch the middleware settings.

    :return dict: settings
    """
    # Get the user-provided settings
    user_settings = dict(getattr(django_settings, _settings_key, {}))
    user_settings_keys = set(user_settings.keys())
    # Check for required but missing settings
    missing = _required_settings_keys - user_settings_keys
    if missing:
        raise AuthzConfigurationError(
            'Missing required {} config: {}'.format(_settings_key, missing))
    # Check for unknown settings
    unknown = user_settings_keys - _available_settings_keys
    if unknown:
        raise AuthzConfigurationError(
            'Unknown {} config params: {}'.format(_settings_key, unknown))
    # Merge defaults with provided settings
    defaults = _available_settings_keys - user_settings_keys
    user_settings.update({key: _available_settings[key] for key in defaults})

    _rectify(user_settings)
    return types.MappingProxyType(user_settings)



"""
    authorization_django.middleware
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
import logging
import sys

from django import http
from django.conf import settings as django_settings
import jwt

from .config import settings


def _create_logger(middleware_settings):
    """ Creates a logger using the given settings.
    """
    if django_settings.DEBUG:
        level = logging.DEBUG
        formatter = logging.Formatter(
            middleware_settings['LOGGER_FORMAT_DEBUG'])
    else:
        level = middleware_settings['LOGGER_LEVEL']
        formatter = logging.Formatter(middleware_settings['LOGGER_FORMAT'])

    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(level)
    handler.setFormatter(formatter)

    logger = logging.getLogger(middleware_settings['LOGGER_NAME'])

    # If in some strange way this logger already exists we make sure to delete
    # its existing handlers
    del logger.handlers[:]

    logger.addHandler(handler)

    # Disable propagation by default
    logger.propagate = False

    return logger


class _AuthorizationHeaderError(Exception):

    def __init__(self, response):
        self.response = response


def authorization_middleware(get_response):
    """ Django middleware to parse incoming access tokens, validate them and
    set an authorization function on the request.

    The decision to use a generic middleware rather than an
    AuthenticationMiddleware is explicitly made, because inctances of the
    latter come with a number of assumptions (i.e. that user.is_authorized()
    exists, or that request.user uses the User model).

    Example usage:

    ::

        request.is_authorized_for()

    :param get_response: callable that creates the response object
    :return: response
    :todo:
        Two things needs to be done when we can completely remove the Django
        JWT plugin:

        - Nested function 'middleware' allows both 'JWT' (not IANA-registered)
          and 'Bearer' as Authorization header prefix; JWT should not be
          accepted.
        - The Django JWT middleware does not include the authz claim, so this
          plugin does not fail if it is not present; this behavior is wrong
          when we no longer use the Django JWT plugin.
    """
    middleware_settings = settings()
    logger = _create_logger(middleware_settings)

    min_scope = middleware_settings['MIN_SCOPE']

    def authorize_function(scopes, token_signature, x_unique_id=None):
        """ Creates a partial around :func:`levels.is_authorized`
        that wraps the current user's scopes.

        :return func:
        """
        log_msg_scopes = 'Granted access (needed: {}, granted: {}, token: {})'

        def is_authorized(*needed_scopes):
            granted_scopes = set(scopes)
            needed_scopes = set(needed_scopes)
            result = needed_scopes.issubset(granted_scopes)
            if result:
                msg = log_msg_scopes.format(needed_scopes, granted_scopes, token_signature)
                if x_unique_id:
                    msg += ' X-Unique-ID: {}'.format(x_unique_id)
                logger.info(msg)
            return result

        return is_authorized

    def authorize_forced_anonymous(_):
        """ Authorize function for routes that are forced anonymous"""
        raise Exception(
            'Should not call is_authorized_for in anonymous routes')

    def insufficient_scope():
        """Returns an HttpResponse object with a 401."""
        msg = 'Bearer realm="datapunt", error="insufficient_scope"'
        response = http.HttpResponse('Unauthorized', status=401)
        response['WWW-Authenticate'] = msg
        return response

    def invalid_token():
        """ Returns an HttpResponse object with a 401
        """
        msg = 'Bearer realm="datapunt", error="invalid_token"'
        response = http.HttpResponse('Unauthorized', status=401)
        response['WWW-Authenticate'] = msg
        return response

    def invalid_request():
        """ Returns an HttpResponse object with a 400
        """
        msg = (
            "Bearer realm=\"datapunt\", error=\"invalid_request\", "
            "error_description=\"Invalid Authorization header format; "
            "should be: 'Bearer [token]'\"")
        response = http.HttpResponse('Bad Request', status=400)
        response['WWW-Authenticate'] = msg
        return response

    def token_data(authorization):
        """ Get the token data present in the given authorization header.
        """
        try:
            prefix, token = authorization.split()
        except ValueError:
            logger.warning(
                'Invalid Authorization header: {}'.format(authorization))
            raise _AuthorizationHeaderError(invalid_request())
        if prefix.lower() != 'bearer':
            logger.warning(
                'Invalid Authorization header: {}'.format(authorization))
            raise _AuthorizationHeaderError(invalid_request())

        try:
            header = jwt.get_unverified_header(token)
        except (jwt.InvalidTokenError, jwt.DecodeError):
            logger.exception("API authz problem: JWT decode error while reading header")
            raise _AuthorizationHeaderError(invalid_token())

        if 'kid' not in header:
            logger.exception("Did not get a valid key identifier")
            raise _AuthorizationHeaderError(invalid_token())

        keys = middleware_settings['JWKS'].verifiers

        if header['kid'] not in keys:
            logger.exception("Unknown key identifier: {}".format(header['kid']))
            raise _AuthorizationHeaderError(invalid_token())

        key = keys[header['kid']]

        try:
            decoded = jwt.decode(token, key=key.key, algorithms=(key.alg,))
        except jwt.InvalidTokenError:
            logger.exception('API authz problem: could not decode access '
                             'token {}'.format(token))
            raise _AuthorizationHeaderError(invalid_token())

        if 'scopes' not in decoded:
            logger.warning('API authz problem: access token misses '
                           'authz and scopes claim: {}'.format(token))
            raise _AuthorizationHeaderError(invalid_token())
        else:
            scopes = decoded['scopes']

        token_signature = token.split('.')[2]
        return scopes, token_signature

    def middleware(request):
        """ Parses the Authorization header, decodes and validates the JWT and
        adds the is_authorized_for function to the request.
        """
        request_path = request.path
        forced_anonymous = any(
            request_path.startswith(route)
            for route in middleware_settings['FORCED_ANONYMOUS_ROUTES'])
        is_options = request.method == 'OPTIONS'

        if forced_anonymous or is_options:
            authz_func = authorize_forced_anonymous
        else:
            authorization = request.META.get('HTTP_AUTHORIZATION')
            token_signature = ''

            if authorization:
                try:
                    scopes, token_signature = token_data(authorization)
                except _AuthorizationHeaderError as e:
                    return e.response
            else:
                scopes = []

            x_unique_id = request.META.get('HTTP_X_UNIQUE_ID')
            authz_func = authorize_function(scopes, token_signature, x_unique_id)

            if len(min_scope) > 0 and not authz_func(min_scope):
                return insufficient_scope()

        request.is_authorized_for = authz_func

        response = get_response(request)

        return response

    return middleware



""" setuptools config
"""
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    """ Custom class to avoid depending on pytest-runner.
    """
    user_options = [('pytest-args=', 'a', "Arguments to pass into py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


with open('README.rst', encoding='utf-8') as f:
    long_description = f.read()

version = '0.2.18'
packages = ['authorization_django']
requires = ['Django>=1.10.3', 'PyJWT>=1.5.2', 'cryptography==2.1.2']
requires_test = ['pytest>=3.0.5', 'pytest-cov>=2.4.0']
requires_extras = {
    'dev': [] + requires_test,
}

setup(
    name='datapunt-authorization-django',
    version=version,
    description='Datapunt authorization check for Django',
    url='https://github.com/Amsterdam/authorization_django',
    author='Amsterdam Datapunt',
    author_email='datapunt.ois@amsterdam.nl',
    license='Mozilla Public License Version 2.0',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Topic :: System :: Systems Administration :: Authentication/Directory',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    cmdclass={'test': PyTest},
    packages=packages,
    install_requires=requires,
    tests_require=requires_test,
    extras_require=requires_extras,
)



import base64
import collections
import json
from types import MappingProxyType

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.utils import int_from_bytes

_Key = collections.namedtuple('Key', 'alg key')
"""Immutable type for key storage"""

_KeySet = collections.namedtuple('KeySet', 'signers verifiers')
"""Immutable type for key sets"""


class JWKError(Exception):
    """Error raised when parsing a JWKSet fails."""


def load(jwks):
    """Parse a JWKSet and return a dictionary that maps key IDs on keys."""
    sign_keys = {}
    verify_keys = {}
    try:
        keyset = json.loads(jwks)
        for key in keyset['keys']:
            for op in key['key_ops']:
                if op == 'sign':
                    k = sign_keys
                elif op == 'verify':
                    k = verify_keys
                else:
                    raise JWKError("Unsupported key operation: {}".format(op))
                if key['kty'] == 'oct':
                    k[key['kid']] = _Key(alg=key['alg'], key=base64.urlsafe_b64decode(key['k']))
                elif key['kty'] == 'EC':
                    alg, ec_key = _load_ecdsa(key, op == 'verify')
                    k[key['kid']] = _Key(alg=alg, key=ec_key)
                else:
                    raise JWKError("Unsupported key type: {}".format(key['kty']))
    except (KeyError, json.JSONDecodeError) as e:
        raise JWKError() from e
    keys = _KeySet(signers=MappingProxyType(sign_keys), verifiers=MappingProxyType(verify_keys))
    return keys


def _load_ecdsa(key, is_verifier):
    if key.get('kty') != 'EC':
        raise JWKError('Not an Elliptic curve key')

    if 'x' not in key or 'y' not in key:
        raise JWKError('Not an Elliptic curve key')

    x = base64.urlsafe_b64decode(key.get('x'))
    y = base64.urlsafe_b64decode(key.get('y'))

    curve = key.get('crv')
    if curve == 'P-256':
        if len(x) == len(y) == 32:
            alg = 'ES256'
            curve_obj = ec.SECP256R1()
        else:
            raise JWKError("Coords should be 32 bytes for curve P-256")
    elif curve == 'P-384':
        if len(x) == len(y) == 48:
            alg = "ES384"
            curve_obj = ec.SECP384R1()
        else:
            raise JWKError("Coords should be 48 bytes for curve P-384")
    elif curve == 'P-521':
        if len(x) == len(y) == 66:
            alg = "ES512"
            curve_obj = ec.SECP521R1()
        else:
            raise JWKError("Coords should be 66 bytes for curve P-521")
    else:
        raise JWKError("Invalid curve: {}".format(curve))

    public_numbers = ec.EllipticCurvePublicNumbers(
        x=int_from_bytes(x, 'big'), y=int_from_bytes(y, 'big'), curve=curve_obj
    )

    if not is_verifier:
        if 'd' not in key:
            raise JWKError("Signing ECDSA keys must contain private key")
        d = base64.urlsafe_b64decode(key.get('d'))
        if len(d) != len(x):
            raise JWKError(
                "D should be {} bytes for curve {}", len(x), curve
            )

        key = ec.EllipticCurvePrivateNumbers(
            int_from_bytes(d, 'big'), public_numbers
        ).private_key(default_backend())
    else:
        key = public_numbers.public_key(default_backend())

    return alg, key
